@@ -4,7 +4,7 @@ set -e
 data_files="UnicodeData.txt CaseFolding.txt EastAsianWidth.txt"
 emoji_files="emoji-data.txt"
 
-UNIDIR_DEFAULT=unicode
+UNIDIR_DEFAULT=src/unicode
 DOWNLOAD_URL_BASE_DEFAULT='http://unicode.org/Public'
 
 if test x$1 = 'x--help' ; then
@@ -30,7 +30,7 @@ for filename in $data_files ; do
 done
 
 for filename in $emoji_files ; do
-  curl -L -o "$UNIDIR/$filename" "$DOWNLOAD_URL_BASE/emoji/latest/$filename"
+  curl -L -o "$UNIDIR/$filename" "$DOWNLOAD_URL_BASE/UNIDATA/emoji/$filename"
   (
     cd "$UNIDIR"
     git add $filename
@@ -39,5 +39,5 @@ done
 
 (
   cd "$UNIDIR"
-  git commit -m "Update unicode files" -- $files
+  git commit -m "feat: update unicode tables" -- $files
 )
