@@ -1,15 +1,16 @@
 #ifndef NVIM_API_PRIVATE_DEFS_H
 #define NVIM_API_PRIVATE_DEFS_H
 
-#include <stdint.h>
 #include <stdbool.h>
+#include <stdint.h>
 #include <string.h>
 
+#include "klib/kvec.h"
 #include "nvim/func_attr.h"
 #include "nvim/types.h"
 
-#define ARRAY_DICT_INIT {.size = 0, .capacity = 0, .items = NULL}
-#define STRING_INIT {.data = NULL, .size = 0}
+#define ARRAY_DICT_INIT KV_INITIAL_VALUE
+#define STRING_INIT { .data = NULL, .size = 0 }
 #define OBJECT_INIT { .type = kObjectTypeNil }
 #define ERROR_INIT { .type = kErrorTypeNone, .msg = NULL }
 #define REMOTE_TYPE(type) typedef handle_T type
@@ -19,13 +20,14 @@
 #ifdef INCLUDE_GENERATED_DECLARATIONS
 # define ArrayOf(...) Array
 # define DictionaryOf(...) Dictionary
+# define Dict(name) KeyDict_##name
 #endif
 
 // Basic types
 typedef enum {
   kErrorTypeNone = -1,
   kErrorTypeException,
-  kErrorTypeValidation
+  kErrorTypeValidation,
 } ErrorType;
 
 typedef enum {
@@ -34,6 +36,7 @@ typedef enum {
   kMessageTypeRequest = 0,
   kMessageTypeResponse = 1,
   kMessageTypeNotification = 2,
+  kMessageTypeRedrawEvent = 3,
 } MessageType;
 
 /// Mask for all internal calls
@@ -83,18 +86,10 @@ REMOTE_TYPE(Window);
 REMOTE_TYPE(Tabpage);
 
 typedef struct object Object;
-
-typedef struct {
-  Object *items;
-  size_t size, capacity;
-} Array;
+typedef kvec_t(Object) Array;
 
 typedef struct key_value_pair KeyValuePair;
-
-typedef struct {
-  KeyValuePair *items;
-  size_t size, capacity;
-} Dictionary;
+typedef kvec_t(KeyValuePair) Dictionary;
 
 typedef enum {
   kObjectTypeNil = 0,
@@ -129,5 +124,14 @@ struct key_value_pair {
   Object value;
 };
 
+typedef Object *(*field_hash)(void *retval, const char *str, size_t len);
+typedef struct {
+  char *str;
+  size_t ptr_off;
+} KeySetLink;
+
+#ifdef INCLUDE_GENERATED_DECLARATIONS
+# include "keysets_defs.generated.h"
+#endif
 
 #endif  // NVIM_API_PRIVATE_DEFS_H
