@@ -3,30 +3,28 @@
 
 #include <assert.h>
 #include <stdint.h>
-
 #include <uv.h>
 
-#include "nvim/event/loop.h"
-#include "nvim/event/socket.h"
-#include "nvim/event/rstream.h"
-#include "nvim/event/wstream.h"
-#include "nvim/os/os.h"
 #include "nvim/ascii.h"
-#include "nvim/vim.h"
-#include "nvim/strings.h"
-#include "nvim/path.h"
+#include "nvim/charset.h"
+#include "nvim/event/loop.h"
+#include "nvim/event/rstream.h"
+#include "nvim/event/socket.h"
+#include "nvim/event/wstream.h"
+#include "nvim/log.h"
+#include "nvim/macros.h"
 #include "nvim/main.h"
 #include "nvim/memory.h"
-#include "nvim/macros.h"
-#include "nvim/charset.h"
-#include "nvim/log.h"
+#include "nvim/os/os.h"
+#include "nvim/path.h"
+#include "nvim/strings.h"
+#include "nvim/vim.h"
 
 #ifdef INCLUDE_GENERATED_DECLARATIONS
 # include "event/socket.c.generated.h"
 #endif
 
-int socket_watcher_init(Loop *loop, SocketWatcher *watcher,
-                        const char *endpoint)
+int socket_watcher_init(Loop *loop, SocketWatcher *watcher, const char *endpoint)
   FUNC_ATTR_NONNULL_ALL
 {
   xstrlcpy(watcher->addr, endpoint, sizeof(watcher->addr));
@@ -40,8 +38,8 @@ int socket_watcher_init(Loop *loop, SocketWatcher *watcher,
     char *port = host_end + 1;
     intmax_t iport;
 
-    int ret = getdigits_safe(&(char_u *){ (char_u *)port }, &iport);
-    if (ret == FAIL || iport < 0 || iport > UINT16_MAX) {
+    int ok = try_getdigits(&(char *){ port }, &iport);
+    if (!ok || iport < 0 || iport > UINT16_MAX) {
       ELOG("Invalid port: %s", port);
       return UV_EINVAL;
     }
@@ -55,10 +53,8 @@ int socket_watcher_init(Loop *loop, SocketWatcher *watcher,
     uv_getaddrinfo_t request;
 
     int retval = uv_getaddrinfo(&loop->uv, &request, NULL, addr, port,
-                                &(struct addrinfo){
-                                  .ai_family = AF_UNSPEC,
-                                  .ai_socktype = SOCK_STREAM,
-                                });
+                                &(struct addrinfo){ .ai_family = AF_UNSPEC,
+                                                    .ai_socktype = SOCK_STREAM, });
     if (retval != 0) {
       ELOG("Host lookup failed: %s", endpoint);
       return retval;
@@ -105,13 +101,12 @@ int socket_watcher_start(SocketWatcher *watcher, int backlog, socket_cb cb)
         // contain 0 in this case, unless uv_tcp_getsockname() is used first.
         uv_tcp_getsockname(&watcher->uv.tcp.handle, (struct sockaddr *)&sas,
                            &(int){ sizeof(sas) });
-        uint16_t port = (uint16_t)(
-            (sas.ss_family == AF_INET)
-            ? (STRUCT_CAST(struct sockaddr_in, &sas))->sin_port
-            : (STRUCT_CAST(struct sockaddr_in6, &sas))->sin6_port);
+        uint16_t port = (uint16_t)((sas.ss_family == AF_INET)
+                                   ? (STRUCT_CAST(struct sockaddr_in, &sas))->sin_port
+                                   : (STRUCT_CAST(struct sockaddr_in6, &sas))->sin6_port);
         // v:servername uses the string from watcher->addr
         size_t len = strlen(watcher->addr);
-        snprintf(watcher->addr+len, sizeof(watcher->addr)-len, ":%" PRIu16,
+        snprintf(watcher->addr + len, sizeof(watcher->addr) - len, ":%" PRIu16,
                  ntohs(port));
         break;
       }
@@ -129,8 +124,8 @@ int socket_watcher_start(SocketWatcher *watcher, int backlog, socket_cb cb)
     if (result == UV_EACCES) {
       // Libuv converts ENOENT to EACCES for Windows compatibility, but if
       // the parent directory does not exist, ENOENT would be more accurate.
-      *path_tail((char_u *)watcher->addr) = NUL;
-      if (!os_path_exists((char_u *)watcher->addr)) {
+      *path_tail(watcher->addr) = NUL;
+      if (!os_path_exists(watcher->addr)) {
         result = UV_ENOENT;
       }
     }
@@ -169,7 +164,7 @@ void socket_watcher_close(SocketWatcher *watcher, socket_close_cb cb)
   FUNC_ATTR_NONNULL_ARG(1)
 {
   watcher->close_cb = cb;
-  uv_close((uv_handle_t *)watcher->stream, close_cb);
+  uv_close(STRUCT_CAST(uv_handle_t, watcher->stream), close_cb);
 }
 
 static void connection_event(void **argv)
@@ -183,7 +178,7 @@ static void connection_cb(uv_stream_t *handle, int status)
 {
   SocketWatcher *watcher = handle->data;
   CREATE_EVENT(watcher->events, connection_event, 2, watcher,
-      (void *)(uintptr_t)status);
+               (void *)(uintptr_t)status);
 }
 
 static void close_cb(uv_handle_t *handle)
@@ -203,9 +198,8 @@ static void connect_cb(uv_connect_t *req, int status)
   }
 }
 
-bool socket_connect(Loop *loop, Stream *stream,
-                    bool is_tcp, const char *address,
-                    int timeout, const char **error)
+bool socket_connect(Loop *loop, Stream *stream, bool is_tcp, const char *address, int timeout,
+                    const char **error)
 {
   bool success = false;
   int status;
@@ -231,7 +225,7 @@ bool socket_connect(Loop *loop, Stream *stream,
                                     .ai_socktype = SOCK_STREAM,
                                     .ai_flags  = AI_NUMERICSERV };
     int retval = uv_getaddrinfo(&loop->uv, &addr_req, NULL,
-                                addr, host_end+1, &hints);
+                                addr, host_end + 1, &hints);
     if (retval != 0) {
       *error = _("failed to lookup host or port");
       goto cleanup;
@@ -243,7 +237,6 @@ tcp_retry:
     uv_tcp_nodelay(tcp, true);
     uv_tcp_connect(&req,  tcp, addrinfo->ai_addr, connect_cb);
     uv_stream = (uv_stream_t *)tcp;
-
   } else {
     uv_pipe_t *pipe = &stream->uv.pipe;
     uv_pipe_init(&loop->uv, pipe, 0);
