@@ -2,6 +2,8 @@
 #define NVIM_FILEIO_H
 
 #include "nvim/buffer_defs.h"
+#include "nvim/eval/typval.h"
+#include "nvim/garray.h"
 #include "nvim/os/os.h"
 
 // Values for readfile() flags
@@ -12,26 +14,15 @@
 #define READ_DUMMY      0x10    // reading into a dummy buffer
 #define READ_KEEP_UNDO  0x20    // keep undo info
 #define READ_FIFO       0x40    // read from fifo or socket
+#define READ_NOWINENTER 0x80    // do not trigger BufWinEnter
+#define READ_NOFILE     0x100   // do not read a file, do trigger BufReadCmd
 
 #define READ_STRING(x, y) (char_u *)read_string((x), (size_t)(y))
 
-/*
- * Struct to save values in before executing autocommands for a buffer that is
- * not the current buffer.
- */
-typedef struct {
-  buf_T       *save_curbuf;     ///< saved curbuf
-  int use_aucmd_win;            ///< using aucmd_win
-  win_T       *save_curwin;     ///< saved curwin
-  win_T       *save_prevwin;    ///< saved prevwin
-  win_T       *new_curwin;      ///< new curwin
-  bufref_T     new_curbuf;      ///< new curbuf
-  char_u      *globaldir;       ///< saved value of globaldir
-} aco_save_T;
+typedef varnumber_T (*CheckItem)(void *expr, const char *name);
 
 #ifdef INCLUDE_GENERATED_DECLARATIONS
 // Events for autocommands
-# include "auevents_enum.generated.h"
 # include "fileio.h.generated.h"
 #endif
 #endif  // NVIM_FILEIO_H
