@@ -1,13 +1,14 @@
 #ifndef NVIM_UI_H
 #define NVIM_UI_H
 
-#include <stddef.h>
 #include <stdbool.h>
+#include <stddef.h>
 #include <stdint.h>
 
-#include "nvim/globals.h"
 #include "nvim/api/private/defs.h"
+#include "nvim/globals.h"
 #include "nvim/highlight_defs.h"
+#include "nvim/memory.h"
 
 typedef enum {
   kUICmdline = 0,
@@ -53,7 +54,12 @@ struct ui_t {
   bool ui_ext[kUIExtCount];  ///< Externalized UI capabilities.
   int width;
   int height;
-  int pum_height;
+  int pum_nlines;  /// actual nr. lines shown in PUM
+  bool pum_pos;  /// UI reports back pum position?
+  double pum_row;
+  double pum_col;
+  double pum_height;
+  double pum_width;
   void *data;
 
 #ifdef INCLUDE_GENERATED_DECLARATIONS
@@ -63,11 +69,16 @@ struct ui_t {
   void (*inspect)(UI *ui, Dictionary *info);
 };
 
+typedef struct ui_event_callback {
+  LuaRef cb;
+  bool ext_widgets[kUIGlobalCount];
+} UIEventCallback;
+
 #ifdef INCLUDE_GENERATED_DECLARATIONS
 # include "ui.h.generated.h"
+
 # include "ui_events_call.h.generated.h"
 #endif
 
-
 EXTERN MultiQueue *resize_events;
 #endif  // NVIM_UI_H
