@@ -50,29 +50,33 @@ local function basic_register_test(noblock)
     text, stuff and some more
     some some text, stuff and some more]])
 
-  -- deleting a word to named ("a) updates "1 (and not "-)
+  -- deleting a word to named ("a) doesn't update "1 or "-
   feed('gg"adwj"1P^"-P')
   expect([[
     , stuff and some more
-    some textsome some text, stuff and some more]])
+    some some random text
+    some some text, stuff and some more]])
 
   -- deleting a line does update ""
   feed('ggdd""P')
   expect([[
     , stuff and some more
-    some textsome some text, stuff and some more]])
+    some some random text
+    some some text, stuff and some more]])
 
   feed('ggw<c-v>jwyggP')
   if noblock then
     expect([[
       stuf
-      me t
+      me s
       , stuff and some more
-      some textsome some text, stuff and some more]])
+      some some random text
+      some some text, stuff and some more]])
   else
     expect([[
       stuf, stuff and some more
-      me tsome textsome some text, stuff and some more]])
+      me ssome some random text
+      some some text, stuff and some more]])
   end
 
   -- pasting in visual does unnamed delete of visual selection
@@ -88,8 +92,13 @@ describe('clipboard', function()
   before_each(function()
     clear()
     screen = Screen.new(72, 4)
+    screen:set_default_attr_ids({
+      [0] = {bold = true, foreground = Screen.colors.Blue},
+      [1] = {foreground = Screen.colors.Grey100, background = Screen.colors.Red},
+      [2] = {bold = true, foreground = Screen.colors.SeaGreen4},
+      [3] = {bold = true, reverse = true};
+    })
     screen:attach()
-    command("set display-=msgsep")
   end)
 
   it('unnamed register works without provider', function()
@@ -103,22 +112,22 @@ describe('clipboard', function()
     feed_command('redir @+> | :silent echo system("cat CONTRIBUTING.md") | redir END')
     screen:expect([[
       ^                                                                        |
-      ~                                                                       |
-      ~                                                                       |
+      {0:~                                                                       }|
+      {0:~                                                                       }|
       clipboard: No provider. Try ":checkhealth" or ":h clipboard".           |
-    ]], nil, {{bold = true, foreground = Screen.colors.Blue}})
+    ]])
   end)
 
   it('`:redir @+>|bogus_cmd|redir END` + invalid g:clipboard must not recurse #7184',
   function()
     command("let g:clipboard = 'bogus'")
     feed_command('redir @+> | bogus_cmd | redir END')
-    screen:expect([[
-      ~                                                                       |
-      clipboard: No provider. Try ":checkhealth" or ":h clipboard".           |
-      E492: Not an editor command: bogus_cmd | redir END                      |
-      Press ENTER or type command to continue^                                 |
-    ]], nil, {{bold = true, foreground = Screen.colors.Blue}})
+    screen:expect{grid=[[
+    {3:                                                                        }|
+    clipboard: No provider. Try ":checkhealth" or ":h clipboard".           |
+    {1:E492: Not an editor command: bogus_cmd | redir END}                      |
+    {2:Press ENTER or type command to continue}^                                 |
+    ]]}
   end)
 
   it('invalid g:clipboard shows hint if :redir is not active', function()
@@ -131,10 +140,10 @@ describe('clipboard', function()
     feed_command('let @+="foo"')
     screen:expect([[
       ^                                                                        |
-      ~                                                                       |
-      ~                                                                       |
+      {0:~                                                                       }|
+      {0:~                                                                       }|
       clipboard: No provider. Try ":checkhealth" or ":h clipboard".           |
-    ]], nil, {{bold = true, foreground = Screen.colors.Blue}})
+    ]])
   end)
 
   it('valid g:clipboard', function()
@@ -148,6 +157,16 @@ describe('clipboard', function()
     eq('', eval('provider#clipboard#Error()'))
   end)
 
+  it('g:clipboard using lists', function()
+    source([[let g:clipboard = {
+            \  'name': 'custom',
+            \  'copy': { '+': ['any', 'command'], '*': ['some', 'other'] },
+            \  'paste': { '+': ['any', 'command'], '*': ['some', 'other'] },
+            \}]])
+    eq('custom', eval('provider#clipboard#Executable()'))
+    eq('', eval('provider#clipboard#Error()'))
+  end)
+
   it('g:clipboard using VimL functions', function()
     -- Implements a fake clipboard provider. cache_enabled is meaningless here.
     source([[let g:clipboard = {
@@ -266,13 +285,17 @@ describe('clipboard (with fake clipboard.vim)', function()
   function()
     local screen = Screen.new(72, 4)
     screen:attach()
+    screen:set_default_attr_ids({
+      [0] = {bold = true, foreground = Screen.colors.Blue},
+      [1] = {foreground = Screen.colors.Grey100, background = Screen.colors.Red},
+    })
     feed_command('redir @+> | bogus_cmd | redir END')
     screen:expect([[
       ^                                                                        |
-      ~                                                                       |
-      ~                                                                       |
-      E492: Not an editor command: bogus_cmd | redir END                      |
-    ]], nil, {{bold = true, foreground = Screen.colors.Blue}})
+      {0:~                                                                       }|
+      {0:~                                                                       }|
+      {1:E492: Not an editor command: bogus_cmd | redir END}                      |
+    ]])
   end)
 
   it('has independent "* and unnamed registers by default', function()
@@ -287,18 +310,18 @@ describe('clipboard (with fake clipboard.vim)', function()
     insert([[
       text:
       first line
-      secound line
+      second line
       third line]])
 
     feed('G"+dd"*dddd"+p"*pp')
     expect([[
       text:
       third line
-      secound line
+      second line
       first line]])
     -- linewise selection should be encoded as an extra newline
     eq({{'third line', ''}, 'V'}, eval("g:test_clip['+']"))
-    eq({{'secound line', ''}, 'V'}, eval("g:test_clip['*']"))
+    eq({{'second line', ''}, 'V'}, eval("g:test_clip['*']"))
   end)
 
   it('handles null bytes when pasting and in getreg', function()
@@ -454,7 +477,7 @@ describe('clipboard (with fake clipboard.vim)', function()
       expect("indeed star")
     end)
 
-    it('unamed operations work even if the provider fails', function()
+    it('unnamed operations work even if the provider fails', function()
       insert('the text')
       feed('yy')
       feed_command("let g:cliperror = 1")
@@ -487,6 +510,20 @@ describe('clipboard (with fake clipboard.vim)', function()
       feed('p')
       eq('textstar', meths.get_current_line())
     end)
+
+    it('Block paste works correctly', function()
+      insert([[
+        aabbcc
+        ddeeff
+      ]])
+      feed('gg^<C-v>') -- Goto start of top line enter visual block mode
+      feed('3ljy^k') -- yank 4x2 block & goto initial location
+      feed('P') -- Paste it infront
+      expect([[
+        aabbaabbcc
+        ddeeddeeff
+      ]])
+    end)
   end)
 
   describe('clipboard=unnamedplus', function()
@@ -522,7 +559,7 @@ describe('clipboard (with fake clipboard.vim)', function()
       eq({{'really unnamed', ''}, 'V'}, eval("g:test_clip['+']"))
       eq({{'really unnamed', ''}, 'V'}, eval("g:test_clip['*']"))
 
-      -- unnamedplus takes predecence when pasting
+      -- unnamedplus takes precedence when pasting
       eq('+', eval('v:register'))
       feed_command("let g:test_clip['+'] = ['the plus','']")
       feed_command("let g:test_clip['*'] = ['the star','']")
@@ -586,10 +623,10 @@ describe('clipboard (with fake clipboard.vim)', function()
       {0:~                                                           }|
       {4:                                                            }|
       :registers                                                  |
-      {1:--- Registers ---}                                           |
-      "*   some{2:^J}star data{2:^J}                                      |
-      "+   such{2:^J}plus{2:^J}stuff                                      |
-      ":   let g:test_clip['+'] = ['such', 'plus', 'stuff']       |
+      {1:Type Name Content}                                           |
+        l  "*   some{2:^J}star data{2:^J}                                 |
+        c  "+   such{2:^J}plus{2:^J}stuff                                 |
+        c  ":   let g:test_clip['+'] = ['such', 'plus', 'stuff']  |
       {3:Press ENTER or type command to continue}^                     |
     ]], {
       [0] = {bold = true, foreground = Screen.colors.Blue},
@@ -620,14 +657,12 @@ describe('clipboard (with fake clipboard.vim)', function()
       '',
       '',
       'E121: Undefined variable: doesnotexist',
-      'E15: Invalid expression: doesnotexist',
     }, 'v'}, eval("g:test_clip['*']"))
     feed_command(':echo "Howdy!"')
     eq({{
       '',
       '',
       'E121: Undefined variable: doesnotexist',
-      'E15: Invalid expression: doesnotexist',
       '',
       'Howdy!',
     }, 'v'}, eval("g:test_clip['*']"))
@@ -637,6 +672,9 @@ describe('clipboard (with fake clipboard.vim)', function()
     feed_command('set mouse=a')
 
     local screen = Screen.new(30, 5)
+    screen:set_default_attr_ids({
+      [0] = {bold = true, foreground = Screen.colors.Blue},
+    })
     screen:attach()
     insert([[
       the source
@@ -646,10 +684,10 @@ describe('clipboard (with fake clipboard.vim)', function()
     screen:expect([[
       the ^source                    |
       a target                      |
-      ~                             |
-      ~                             |
+      {0:~                             }|
+      {0:~                             }|
                                     |
-    ]], nil, {{bold = true, foreground = Screen.colors.Blue}})
+    ]])
 
     feed('<MiddleMouse><0,1>')
     expect([[
