@@ -2,6 +2,8 @@
 local helpers = require('test.functional.helpers')(after_each)
 local nvim_command, funcs, meths, nvim_feed, eq =
   helpers.command, helpers.funcs, helpers.meths, helpers.feed, helpers.eq
+local assert_alive = helpers.assert_alive
+local expect_exit = helpers.expect_exit
 
 local shada_helpers = require('test.functional.shada.helpers')
 local reset, clear = shada_helpers.reset, shada_helpers.clear
@@ -24,13 +26,13 @@ describe('ShaDa support code', function()
     nvim_command('set shada=\'0 history=2')
     nvim_feed(':" Test\n')
     nvim_feed(':" Test 2\n')
-    nvim_command('qall')
+    expect_exit(nvim_command, 'qall')
     reset()
     nvim_command('set shada=\'0 history=2')
     nvim_command('rshada')
     eq('" Test 2', funcs.histget(':', -1))
     eq('" Test', funcs.histget(':', -2))
-    nvim_command('qall')
+    expect_exit(nvim_command, 'qall')
   end)
 
   it('respects &history when dumping',
@@ -137,7 +139,7 @@ describe('ShaDa support code', function()
     nvim_command('set hlsearch shada-=h')
     nvim_feed('/test\n')
     eq(1, meths.get_vvar('hlsearch'))
-    nvim_command('qall')
+    expect_exit(nvim_command, 'qall')
     reset()
     eq(1, meths.get_vvar('hlsearch'))
   end)
@@ -146,7 +148,7 @@ describe('ShaDa support code', function()
     nvim_command('set hlsearch shada-=h')
     nvim_feed('/test\n')
     nvim_command('nohlsearch')
-    nvim_command('qall')
+    expect_exit(nvim_command, 'qall')
     reset()
     eq(0, meths.get_vvar('hlsearch'))
   end)
@@ -155,7 +157,7 @@ describe('ShaDa support code', function()
     nvim_command('set hlsearch')
     nvim_feed('/test\n')
     eq(1, meths.get_vvar('hlsearch'))
-    nvim_command('qall')
+    expect_exit(nvim_command, 'qall')
     reset()
     eq(0, meths.get_vvar('hlsearch'))
   end)
@@ -174,7 +176,7 @@ describe('ShaDa support code', function()
   it('dumps and loads history with UTF-8 characters', function()
     reset()
     nvim_feed(':echo "«"\n')
-    nvim_command('qall')
+    expect_exit(nvim_command, 'qall')
     reset()
     eq('echo "«"', funcs.histget(':', -1))
   end)
@@ -182,7 +184,7 @@ describe('ShaDa support code', function()
   it('dumps and loads replacement with UTF-8 characters',
   function()
     nvim_command('substitute/./«/ge')
-    nvim_command('qall!')
+    expect_exit(nvim_command, 'qall!')
     reset()
     funcs.setline('.', {'.'})
     nvim_command('&')
@@ -192,7 +194,7 @@ describe('ShaDa support code', function()
   it('dumps and loads substitute pattern with UTF-8 characters',
   function()
     nvim_command('substitute/«/./ge')
-    nvim_command('qall!')
+    expect_exit(nvim_command, 'qall!')
     reset()
     funcs.setline('.', {'«\171'})
     nvim_command('&')
@@ -203,7 +205,7 @@ describe('ShaDa support code', function()
   function()
     nvim_command('silent! /«/')
     nvim_command('set shada+=/0')
-    nvim_command('qall!')
+    expect_exit(nvim_command, 'qall!')
     reset()
     funcs.setline('.', {'\171«'})
     nvim_command('~&')
@@ -216,7 +218,7 @@ describe('ShaDa support code', function()
     -- \171 is U+00AB LEFT-POINTING DOUBLE ANGLE QUOTATION MARK in latin1
     nvim_command('silent! /\171/')
     nvim_command('set shada+=/0')
-    nvim_command('qall!')
+    expect_exit(nvim_command, 'qall!')
     reset()
     funcs.setline('.', {'\171«'})
     nvim_command('~&')
@@ -237,4 +239,13 @@ describe('ShaDa support code', function()
     nvim_command('wshada')
   end)
 
+  it('does not crash when number of history save to zero (#11497)', function()
+    nvim_command('set shada=\'10')
+    nvim_feed(':" Test\n')
+    nvim_command('wshada')
+    nvim_command('set shada=\'10,:0')
+    nvim_command('wshada')
+    assert_alive()
+  end)
+
 end)
